/*
Licensed Materials - Property of IBM
5725-C95
(C) Copyright IBM Corporation 2012. All Rights Reserved.
US Government Users Restricted Rights- Use, duplication or disclosure
restricted by GSA ADP Schedule Contract with IBM Corp.
*/

// coach data stateful class js

define(["dojo/_base/declare",
        "dojo/_base/array",
        "dojo/_base/lang",
        "dojo/_base/json",
        "dojo/Stateful",
        "require",
        "dojox/date/hebrew/Date",
        "dojox/date/islamic/Date",
        "./commonHelper",
        "./dataBindingHelper",
        "./List",
        "./controller",
        "./engine",
        "./logger"
], function(declare, array, lang, json, stateful, require, hebrewDate, islamicDate, commonHelper, dataBindingHelper, list, controller, engine, logger) {
	//Extends dojo.Stateful.  Replaces the "get" method to handle wrapping embedded objects and lists to also make them
	//watchable.  It also caches the watchable objects so that subsequent get operations do not need to wrap new objects.
	return declare("com.ibm.bpm.coach.Stateful", [stateful], {
		//cache of all the child objects and lists
		childrenCache: null,
		_systemCallbackHandle: null,

		//constructor
		constructor: function(baseObject, path) {
			this.inherited(arguments);
			this.childrenCache = {};
			this._objectPath = path;
			this._systemCallbackHandle = null;
			this.enableSystemNotifications();
		},

		disableSystemNotifications: function () {
			this._systemCallbackHandle.unwatch()
		},
		
		enableSystemNotifications: function () {
			//set system call back, for notifications to controller
			this._systemCallbackHandle = this.registerCallback(null, lang.hitch(this, function(prop, oldVal, newVal) {
				    						controller.systemChangeHandler(this._objectPath, prop, newVal);
				    					})
			);			
		},
		
		//overide the super's get to add caching and wrapping of embeded objects and lists
		get: function (name) {
			logger.log("Stateful", "get", "Getting Property: " + this._objectPath + "." + name, logger.FINER);
			var result;
			
			var pathArray = name.split(".");
			if (pathArray.length > 1) {
				result = this.resolvePath(pathArray);
			} else {
				//first check the cache
				var openIndex = name.indexOf("["); 
				if (openIndex != -1) {
					//refers to an array
					var closeIndex = name.indexOf("]");
					logger.log("Stateful", "get", "Getting list index: " + name.substring(openIndex+1, closeIndex), logger.FINER);
					result = this.get(name.substring(0,openIndex)).get(name.substring(openIndex+1, closeIndex));
					return result;
				}
				var property = this.childrenCache[name];
	
				if (typeof(property) == "undefined") {
					var obj = this.inherited(arguments);
					if ((typeof obj == "undefined") || obj == null ){
						return obj;
					}
					//check if it is a simple type
					var dataTypes = ["number", "string", "boolean"];
					array.forEach(dataTypes, function(dataType) {
						if (typeof obj == dataType) {
							result = obj;
						}
					});
	
					if (typeof result == "undefined") {
						//handle a list
						if (Object.prototype.toString.call( obj ) === "[object Array]" || !!obj && !!obj.items && !!obj.listAllSelectedIndices
							//defect 121645 - lists returned from AJAX services use "selected" instead of listAllSelectedIndecies
						    || (!!obj && !!obj["@metadata"] && !!obj["items"] && !!obj["selected"])
						    ) {
							var objectPath = this._objectPath;
							require(["./List"], function(list) {
								result = new list(obj,  objectPath + "." + name);
							});
							//add result to the cache
							this.childrenCache[name] = result;
			        	} 
						else if (typeof obj == "object") {
				        			//handle an object
								if (obj == null || !obj.prototype) { // check that this is a literal
									result = obj;
								} else if (obj.constructor) {
									result = new obj.constructor();
								}
								
								//treat Date like  primitive type
								if (!(result instanceof Date)) {
									result = new com.ibm.bpm.coach.Stateful(result, this._objectPath + "." + name);
			
									//add result to the cache
									this.childrenCache[name] = result;
								}
						}
					} 
					else {
						//handle a simple type
						result = obj;
					}
				} else {
					//cache hit
					result = property;
				}
			}
			return result;
		}, //function get
		
		//Override the super's set to add path
		set: function (name, value, disableSystemNotification, triggeredByConstructor) {
			// dojo Statful calls set() during the execution of its constructor in version 1.8.3.
			// If first parameter (i.e. name) is an object, we know set() is called by the constructor of dojo Stateful,
			if (!!name){			
				// If an object is used, iterate through object
				if(typeof name === "object"){
					for(var x in name){
						if(name.hasOwnProperty(x) && x !="_watchCallbacks"){
							// we know set() is called by the constructor of dojo Stateful
							this.set(x, name[x], true, true);
						}
					}							
					return;
				}
			} else {
				return;
			}		
		
			logger.log("Stateful", "set", "Setting object: " + this._objectPath + "." + name + "= " + value, logger.FINE);
			
			// check if it makes sense to accept the set request
			var currentPlainValue = dataBindingHelper.getPlainValue(this, name);
			if(commonHelper.isValuesEqual(currentPlainValue, value) == commonHelper.YES) {
				// new value is equal to current value, ignore this set request
				logger.log("Stateful", "set", "Ignore this set request because new value is equal to current value. value = " + currentPlainValue, logger.FINE);
				return;
			}
			
			var pathArray = name.split(".");
			if (pathArray.length > 1) {
				name = pathArray.pop();
				var object = this.resolvePath(pathArray);
				if (object instanceof com.ibm.bpm.coach.Stateful) {
					object.set(name, value, disableSystemNotification, triggeredByConstructor);
				}
			} else {
				//defect where a stateful object is being set, need to serialize it and recreate it first
				if (value instanceof com.ibm.bpm.coach.Stateful || value instanceof com.ibm.bpm.coach.List) {
					value = json.fromJson(value.toJson());
				}
			    //test if this is a date and if so turn it into a date object
			    //dates will look like the following yy-mm-ddThh:mm:ss
			    if (dataBindingHelper.isDateString(""+value))
			    {
			    	//it is a date
			    	this.set(name, dataBindingHelper.stringToDate(value), disableSystemNotification);
			    }
			    else {
                    // support for Hebrew and Islamic dates
                    if (value instanceof islamicDate || value instanceof hebrewDate) {
                        value = value.toGregorian();
                    }
					
				    var oldWrappedValue = this.childrenCache[name];
				    
				    //check if the value has changed.  If it has not, do nothing. oldWrappedValue is undefined for simple type, the new value could be undefined
				    if ((value instanceof Date) || ((value != oldWrappedValue) || (typeof oldWrappedValue == "undefined")))
				    {				    
						delete this.childrenCache[name];
						if (oldWrappedValue instanceof com.ibm.bpm.coach.List){
							value = oldWrappedValue.duplicate(value);
						}
						if (disableSystemNotification && disableSystemNotification == true) {
							this.disableSystemNotifications();
							this.inherited(arguments);
							this.enableSystemNotifications();
						} else {
							this.inherited(arguments);
						}						
						if(!triggeredByConstructor) {
							// if set() is triggered by constructor, here is what happens:
							// when stateful object "local.a" is constructed, set() is called on "local.a.b" etc. If dataBindingHelper.updatePlainValueInStatefulChain("local.a", "b", valueB).
							// dataBindingHelper.updatePlainValueInStatefulChain() tries to get Stateful object "local.a" by calling get(). Since stateful object "local.a" doesn't exist yet (not fully constructed),
							// another Stateful object for "local.a" starts to construct. As the result, endless recursive call occurs.
							try {
								dataBindingHelper.updatePlainValueInStatefulChain(this._objectPath, name, this[name]);
							}catch(ex) {
	    						var errorMsg = "Stateful.set() caught Exception when calling dataBindingHelper.updatePlainValueInStatefulChain(). _objectPath is " + this._objectPath + ", name is " + name + ", Exception: " + ex;
	    						commonHelper.consoleLog(errorMsg, commonHelper.ERROR);
							}
						}
				   }
			    }
			}
		},

		//walks an array of strings that reprsent an object graph of Stateful objects
		//returns the leaf or null if the path is not valid
		resolvePath: function(objectArray) {
			var result = null;
			if (objectArray) {
				result = this.resolveIndexedProperty(objectArray[0]);					
				
				for (var i=1; i<objectArray.length; i++) {
					if (result instanceof com.ibm.bpm.coach.Stateful) {
						result = result.resolveIndexedProperty(objectArray[i]);
					} else {
						result = null;
						break;
					}
				}
			}
			return result;
		},
		
		resolveIndexedProperty: function (property) {
			var result = null
			var openIndex = property.indexOf("["); 
			if (openIndex != -1) {
				//refers to an array
				var closeIndex = property.indexOf("]");
				result = this.get(property.substring(0,openIndex)).get(property.substring(openIndex+1, closeIndex));
			} else {
				result = this.get(property);					
			}
			return result;
		},

		//registers a callback for this object that will be invoked if this object is modified
		registerCallback: function (property, callback) {
			var handle;
			if (callback) {
				if (property) {
					//com.ibm.bpm.coach.GlobalAssets["watchHandles"].push(this.watch(property, callback));
					handle = this.watch(property, callback);
				} else {
					//com.ibm.bpm.coach.GlobalAssets["watchHandles"].push(this.watch(callback));
					handle =  this.watch(callback);
				}
				handle.unbind = function (){						
					this.unwatch();
				}
			}
			return handle;
		}, 
		
	    bind: function(property, callback, scope) {
	    	var handle;
	    	var obj = this;
	    	if (property) {
				var pathArray = property.split(".");
				if (pathArray.length > 1) {
					property = pathArray.pop();
					obj = this.resolvePath(pathArray);
				}
	    	}    	
			if (obj instanceof com.ibm.bpm.coach.Stateful) {
		    	handle = obj.registerCallback(property, function(object, oldVal, newVal) {
					//Change handler for Stateful objects, calls the subclass change handler
					var event = {property: object, oldVal: oldVal, newVal: newVal};
					try {
						callback.call(scope, event);
					} catch(ex) {
						commonHelper.consoleLog(ex, commonHelper.ERROR);
					}
				});
			}
			return handle;
	    },

	    bindAll: function(callback, scope) {
	    	return this.bind(null, callback, scope);
	    },
	    
		toJson: function() {
			logger.log("Stateful", "toJson", "Serializing object: " + this._objectPath, logger.FINER);
			//first get all of the stateful properties and do a toJson on them and 
			//then print the simple properties
			var excludeList = ["_inherited", "childrenCache", "_objectPath", "declaredClass", "_systemCallbackHandle"];

			var resultList = [];
			for (var x in this.childrenCache) {
                var jsonValue = this.get(x) != null ? this.get(x).toJson() : null;
				resultList.push("\"" + x + "\":" + jsonValue);
				excludeList.push(x);
			}
			for (var x in this) {
				if (typeof this[x] != 'function'){
					// Only need to handle the properties that are owned by this object, ignore the inherited properties.
					if(!this.hasOwnProperty(x)) {
						continue;
					}

					var skipVar = false;
					for (var j=0; j<excludeList.length; j++) {
						if (excludeList[j] == x) {
							skipVar = true;
							break;
						}
					}
					if (!skipVar) {
						if (typeof this[x] == 'object' ||
							Object.prototype.toString.call(this[x] ) === "[object Array]"
							) {
							if (this[x] instanceof Date) {
								var dateString = this.dateToString(this[x]);								
								resultList.push("\"" + x + "\":" + "\"" + dateString + "\"");
							} else {
                                var jsonValue = this.get(x) != null ? this.get(x).toJson() : null;
								resultList.push("\"" + x + "\":" + jsonValue);
							}
						} else {
							resultList.push("\"" + x + "\":" + json.toJson(this[x]));
						}
					}
				}
			}
			var result = "{" + resultList.join(",") + "}"; 
			logger.log("Stateful", "toJson", result, logger.FINER);
			return result;
		},
		
		stringToDate: function (dateStr) {
			// Refactored this function's logic to a common function but left
			// this function in place just in case any custom code used it.
			return dataBindingHelper.stringToDate(dateStr);
		}, 
		
		dateToString: function (localDate) {
			var date = new Date(localDate.getUTCFullYear(), localDate.getUTCMonth(), localDate.getUTCDate(), localDate.getUTCHours(), localDate.getUTCMinutes(), localDate.getUTCSeconds(), localDate.getUTCMilliseconds());
            
            var yearString = null;
            var yearInt = parseInt(date.getFullYear());
            if (yearInt < 10) {
                yearString = "000" + yearInt
            } else if (yearInt < 100) {
                yearString = "00" + yearInt
            } else if (yearInt < 1000) {
                yearString = "0" + yearInt
            } else {
                yearString = "" + yearInt;
            }
            
			var monthString = null;
			var monthInt = parseInt(date.getMonth()) +1;
			(monthInt <10) ? monthString = "0" + monthInt : monthString = "" + monthInt;

            var dayString = null;
            var dayInt = parseInt(date.getDate());
            (dayInt <10) ? dayString = "0" + dayInt : dayString = "" + dayInt;

            var hourString = null;
            var hourInt = parseInt(date.getHours());
            (hourInt <10) ? hourString = "0" + hourInt : hourString = "" + hourInt;
                        
            var minString = null;
            var minInt = parseInt(date.getMinutes());
            (minInt <10) ? minString = "0" + minInt : minString = "" + minInt;

            var secString = null;
            var secInt = parseInt(date.getSeconds());
            (secInt <10) ? secString = "0" + secInt : secString = "" + secInt;
            
            var millisecString = null;
            var millisecInt = parseInt(date.getMilliseconds());
            (millisecInt < 10) ? 
                millisecString = "00" + millisecInt : 
                (millisecInt < 100) ? 
                    millisecString = "0" + millisecInt : 
                    millisecString = "" + millisecInt;
                                    
            var dateString = yearString + '-' + monthString + '-' + dayString;
            dateString += "T" + hourString + ':' + minString + ':' + secString + "." + millisecString + "Z";
            
            return dateString;      
        }

  	}); //com.ibm.bpm.coach.Stateful	
});

