/*
Licensed Materials - Property of IBM
5725-C95
 Copyright IBM Corporation 2012. All Rights Reserved.
US Government Users Restricted Rights- Use, duplication or disclosure
restricted by GSA ADP Schedule Contract with IBM Corp.
*/

// Contains all common utilities

define(["dojo/_base/lang", "dojo/_base/array", "./CONSTANTS", "./commonHelper", "./bindingResolver", "./dynamicBindingHelper"], function(lang, array, CONSTANTS, commonHelper, bindingResolver, dynamicBindingHelper) {
				
	var specificBindingHelper = new Object();
	
	/*	
	 Specific bind
	 example of bindingStr: 
	 "local.name", "local.contactBook.owner", 
	 "local.contactBook.phones", "local.contactBook.contacts[2]", "local.contactBook.contacts[1].name", 
	 "local.contactBook.contacts[1].phones[]", "local.contactBook.contacts[1].phones[2].number",
	 "local.contactBook.contacts.listSelected", "local.contactBook.contacts.listSelected.phones.listSelected.name", "local.contactBook.contacts.listSelected.phones[0].name",
	 "config.config1.myOption", "config.config2._metadata.label", "config.config3.myOption.p1",
	 "config.config3.myListOption[1]", "config.config3.myListOption[1].p1",
	 "config.config3.myListOption.listSelected", "config.config3.myListOption.listSelected.p2"
	 
	 "local.contactBook.phones[]" is acceptable, it is interpreted as the recommended notation "local.contactBook.phones". Those 2 notations mean that the change on list itself.
	 
	  When callback is invoked, input parameter event object is expected to be readonly, should not be modified by callback function.
	*/
	specificBindingHelper.bind = function(specificBindingStr, callback, scope) {
		if(!specificBindingStr || !callback) {
			return;
		}
				
		var sourcePathArray = specificBindingStr.split(".");
		
		// only accept the binding path that starts with "config" or "local". 
		// For example, some binding paths may start with "resource", which is supposed to be readonly. Make no sense to listen to the change which never occurs.
		if(sourcePathArray[0] != "config" && sourcePathArray[0] != "local") {			
			return;
		}
		
		if(!!scope) {
			callback = lang.hitch(scope, callback); 
		}

		if(sourcePathArray[0] == "config") {
			var isMetadata = (sourcePathArray[2] == "_metadata" && sourcePathArray.length == 4);
			if((sourcePathArray.length == 3 && sourcePathArray[2].indexOf("[") == -1) || isMetadata) {
				// those options are top level, so no need to hook up them on dynamicBindingHelper for the structure change
				var resolvedVariable = bindingResolver.resolveBinding(specificBindingStr);
				
				if(!!resolvedVariable.property) {
					// for "config", resolvedVariable.property is "value". property = "value" doesn't make sense in the callback event.
					// so we need to pass the real property for the callback event.
					var realProperty = isMetadata ? (sourcePathArray[2] + "." + sourcePathArray[3]) : sourcePathArray[2];
					return resolvedVariable.boundObject.bind(resolvedVariable.property, function(rawEvent) {
						var event = {property: realProperty, oldVal: rawEvent.oldVal, newVal: rawEvent.newVal};
						// the event that is from dynamicBindingHelper always contains "path" property. So prepare "path" here to be consistent.
						event.path = specificBindingStr;
						callback(event);
					});
				}
			}
		}
		
		// lastPath that ends with "[]" (e.g. "local.contactBook.phones[]") is translated as bare list (e.g. "local.contactBook.phones")
		var lastPath = sourcePathArray[sourcePathArray.length - 1];
		if(lastPath.indexOf("[]") != -1) {
			sourcePathArray[sourcePathArray.length - 1] = lastPath.replace("[]", "");
			specificBindingStr = sourcePathArray.join(".");
		}
	
		var generalizedBindingStr = specificBindingStr.replace(/\[\d*\]/g, CONSTANTS.INDEX_WITH_BRACKET);
		if(generalizedBindingStr != specificBindingStr) {
			// 
			var generalizedBindingInfo = specificBindingHelper.generalizedBindingStrMap[generalizedBindingStr];
			if(!generalizedBindingInfo) {
				var generalizedCallbackFunc = function(event) {
					// "this" is generalizedBindingInfo object
					var specificCallbackArray = this.specificBindingStrToCallbackMap[event.path];
					if(specificCallbackArray) {
						for(var k=0; k<specificCallbackArray.length; k++) {
							try {
								specificCallbackArray[k](event);
							} catch(ex) {
								var errorMsg = "specificBindingHelper caught Exception when invoking callback function for change on " + event.path + ", Exception: " + ex;
								commonHelper.consoleLog(errorMsg, commonHelper.ERROR);
							}
						}
					}
				}
				
				generalizedBindingInfo = {
					"specificBindingStrToCallbackMap": {}
				}
				generalizedBindingInfo.bindingHandler = dynamicBindingHelper.bind(generalizedBindingStr, generalizedCallbackFunc, generalizedBindingInfo);
								
				specificBindingHelper.generalizedBindingStrMap[generalizedBindingStr] = generalizedBindingInfo;
			}
			
			var callbackList = generalizedBindingInfo.specificBindingStrToCallbackMap[specificBindingStr];
			if(!callbackList) {
				callbackList = [];
				generalizedBindingInfo.specificBindingStrToCallbackMap[specificBindingStr] = callbackList;
			}
			callbackList.push(callback);
			
			var handle = {
				"_type": CONSTANTS.HANDLE_TYPE_SPECIFIC,
				"_generalizedBindingStr": generalizedBindingStr,
				"bindingStr": specificBindingStr,
				"callback": callback,
				"unbind": function() {
					specificBindingHelper.unbindOne(handle);
				}
			}			
			return handle;
		} else {
			return dynamicBindingHelper.bind(specificBindingStr, callback);
		}
	}
	
	// perform better than unbindOne() in general
	specificBindingHelper.unbindMultiple = function(handleArray) {
		if(!!handleArray) {
			var dynamicBindingHandleArray = [];
			for(var k=0; k<handleArray.length; k++) {
				var handle = handleArray[k];
				if(handle._type == CONSTANTS.HANDLE_TYPE_SPECIFIC) {
					var dynamicBindingHandle = specificBindingHelper_removeCallback(handle);
					if(dynamicBindingHandle) {
						dynamicBindingHandleArray.push(dynamicBindingHandle);
					}
				} else if(handle._type == CONSTANTS.HANDLE_TYPE_DYNAMIC) {
					dynamicBindingHandleArray.push(handle);
				} else {
					// assume the type is CONSTANTS.HANDLE_TYPE_NATIVE			
					handleArray[k].unbind();
				}		
			}
			
			if(dynamicBindingHandleArray.length > 0) {
				dynamicBindingHelper.unbindMultiple(dynamicBindingHandleArray);
			}
		}
	}
	
	specificBindingHelper.unbindOne = function(handle) {
		if(handle._type == CONSTANTS.HANDLE_TYPE_SPECIFIC) {
			var dynamicBindingHandle = specificBindingHelper_removeCallback(handle);
			if(dynamicBindingHandle) {
				dynamicBindingHelper.unbindOne(dynamicBindingHandle);
			}
		} else if(handle._type == CONSTANTS.HANDLE_TYPE_DYNAMIC) {
			dynamicBindingHelper.unbindOne(handle);
		} else {
			// assume the type is CONSTANTS.HANDLE_TYPE_NATIVE			
			handle.unbind();
		}
	}
	
	specificBindingHelper_removeCallback = function(handle) {
		var generalizedBindingInfo = specificBindingHelper.generalizedBindingStrMap[handle._generalizedBindingStr];
		if(generalizedBindingInfo) {
			var callbackArray = generalizedBindingInfo.specificBindingStrToCallbackMap[handle.bindingStr];
			if(callbackArray) {
				var index = array.indexOf(callbackArray, handle.callback);
				if(index != -1) {
					callbackArray.splice(index, 1);
					if(callbackArray.length == 0) {
						generalizedBindingInfo.specificBindingStrToCallbackMap[handle.bindingStr] = null;
						delete generalizedBindingInfo.specificBindingStrToCallbackMap[handle.bindingStr];
						
						if(commonHelper.isMapEmpty(generalizedBindingInfo.specificBindingStrToCallbackMap)) {
							specificBindingHelper.generalizedBindingStrMap[handle._generalizedBindingStr] = null;
							delete specificBindingHelper.generalizedBindingStrMap[handle._generalizedBindingStr];
							return generalizedBindingInfo.bindingHandler;
						}
					}
				}
			}
		}
		return null;
	}
	
	specificBindingHelper.generalizedBindingStrMap = {};
	
	return specificBindingHelper;
});
