/*
Licensed Materials - Property of IBM
5725-C95
(C) Copyright IBM Corporation 2013. All Rights Reserved.
US Government Users Restricted Rights- Use, duplication or disclosure
restricted by GSA ADP Schedule Contract with IBM Corp.
*/

// 

define(["dojo/_base/lang"], function(lang) {
				
	var templateStore = {typeToResolvedTemplateMap: {}, typeToCycledTemplateMap: {}, typeToRawTemplateMap: {}, pathToTypeMap: {}};

	templateStore.init = function(globalAssets) {
		// dataStructures.defaults typically has content like following:
		// defaults:{S:"",I:0,D:0.0,DT:new Date(),B:false}
		if(globalAssets.dataStructures) {
			templateStore.typeToResolvedTemplateMap = globalAssets.dataStructures.defaults;
			templateStore.typeToRawTemplateMap = globalAssets.dataStructures.templates;
			// delete the reference in globalAssets
			delete globalAssets.dataStructures;
		}
	}
	
	templateStore.getByType = function(type) {
		var resolvedTemplate = templateStore.typeToResolvedTemplateMap[type];
		// Note: "" and 0 are valid template
		if(resolvedTemplate === undefined) {
			resolvedTemplate = templateStore.typeToCycledTemplateMap[type];
			if(resolvedTemplate === undefined) {
				// the resolved template for this type will be saved into templateStore.typeToResolvedTemplateMap or templateStore.typeToCycledTemplateMap
				// during the execution of templateStore._assembleTemplate.
				var assembleResult = templateStore._assembleTemplate(type, true, {});
				resolvedTemplate = assembleResult.template; 
			}
		}
		return resolvedTemplate;
	}
	
	templateStore.getTypeByPath = function(path) {
		var refinedPath = templateStore._generalizePath(path);
		return templateStore._getTypeByPath(refinedPath);
	}
	
	// input parameter path should be generalized one
	templateStore._getTypeByPath = function(path, pathArray) {
		if(templateStore.pathToTypeMap[path]) {
			return templateStore.pathToTypeMap[path];
		} else {
			// try to figure it out my type from my path information
			if(!pathArray) {
				pathArray = path.split(".");
			}
			if(pathArray.length > 1) {
				var lastPath = pathArray[pathArray.length - 1];
				var newPathArray = pathArray.slice(0, (pathArray.length - 1));
				var newPath = newPathArray.join(".");
				var parentType = templateStore._getTypeByPath(newPath, newPathArray);
				if(parentType) {
					var parentRawTemplate = templateStore.typeToRawTemplateMap[parentType];
					if(parentRawTemplate && parentRawTemplate[lastPath]) {
						var type = parentRawTemplate[lastPath].replace("[]", "");
						templateStore.pathToTypeMap[path] = type;
						return type;
					}
				}
			}
		}
		return null;		
	}
	
	templateStore.save = function(path, type) {
		var refinedPath = templateStore._generalizePath(path);
		if(!templateStore.pathToTypeMap[refinedPath]) {
			templateStore.pathToTypeMap[refinedPath] = type;
		}		
	}
	
	templateStore._generalizePath = function(path) {
		return path.replace(/\[.*\]/g, "");
	}
					
	// assemble template for specified type.
	// Note: templateStore.typeToResolvedTemplateMap and templateStore.typeToCycledTemplateMacan only be edited in this function after they are initialized
	templateStore._assembleTemplate = function(type, isStartPoint, typeToInterimTemplateMap) {
		var result = {template: {}, isCycled: false};
		var rawTemplate = templateStore.typeToRawTemplateMap[type];
		
		if(!rawTemplate) {
			// if rawTemplate doesn't exist, we assume it was deleted when it was visited previously. Now it's visited again, so cycled reference occurs
			result.isCycled = true;
			typeToInterimTemplateMap[type] = result.template; // value is blank object {}
		} else {
			// delete templateStore.typeToRawTemplateMap[type] to avoid endless loop of type referencing.
			// e.g. initially templateStore.typeToRawTemplateMap is 
			// {t3: {p1: "S", p2: "t2", p3: "t2"}, t2: {p4: "I[]", p5: "t1", p6: "t2"}, t1: {p7: "DT", p8: "t3", p9: "t2"}}
			// based on current algorithm (depth firstly traverse), possible endless loop are on following properties in order: p8, p9, p6
			// However, since templateStore.typeToRawTemplateMap["t3"] has been deleted, p8 gets {} as its template. So are p9 and p6
			// As to p3, when it is visited, templateStore.typeToResolvedTemplateMap["t2"] has been created thanks to p2, so no issue.
			delete templateStore.typeToRawTemplateMap[type];
			
			for(var property in rawTemplate) {
				var subType = rawTemplate[property];
				if(subType.indexOf("[]") != -1) {
					// List type, create an empty List with proper te (template)
					result.template[property] = {items: [], listAllSelectedIndices: [], te: subType.replace("[]", "")};
				} else {
					var subTemplate = templateStore.typeToResolvedTemplateMap[subType];
					// Note: "" and 0 are valid template
					if(subTemplate === undefined) {
						subTemplate = typeToInterimTemplateMap[subType];
						if(subTemplate === undefined) {
							var subAssembledResult = templateStore._assembleTemplate(subType, false, typeToInterimTemplateMap);
							// subAssembledResult.template exists for sure because the worst case is that templateStore._assembleTemplate() provides a blank object
							subTemplate = subAssembledResult.template;
							if(subAssembledResult.isCycled) {
								result.isCycled = true;
							}
						} else {
							// subTemplate comes from typeToInterimTemplateMap, all the templates in typeToInterimTemplateMap has cycled reference
							result.isCycled = true;
						}
					}
					result.template[property] = lang.clone(subTemplate);
				}
			}
			
			if(result.isCycled) {
				// the resolved template is in favour of the starting type. so we only save the starting type in templateStore.typeToCycledTemplateMap, 
				// put others in typeToInterimTemplateMap.
				// e.g. {t1: {p1: t2, p2: t2}, t2: {p3: t1}}, if starting type is t1, the resolved result would be
				// t1 -- {p1: {p3: {}}, p2: {p3: {}}}
				// t2 -- {p3: {}}
				// t1 gets its best version, the template goes to templateStore.typeToCycledTemplateMap. 
				// t2's current template is in favour of t1, so it goes to typeToInterimTemplateMap which only exists when t1 is assembled.
				// when t2 becomes starting type, it will get its final version, i.e. t2  -- {p3: {p1: {}, p2: {}}}				
				if(isStartPoint) {
					templateStore.typeToCycledTemplateMap[type] = result.template;
				} else {
					// don't check if typeToInterimTemplateMap[type] already exists, simply update the interim template even if it exists. 
					// See following example for the reason:
					// {t1: {p1: t2, p2: t1, p3: t2}, t2: {p4: t2, p5: t1}}
					// first interim template for t2 is blank object {}, created when p4 is visited. interim template for t1 is also blank object {} 
					// when p5 is visited.
					// when p1 is completed, new interim template for t2 is worked out as {p4: {}, p5{}}. p2 still use {} as interim template for t1.
					// when p3 is visited, t2's new interim template is used.
					// final template for t1 is {p1: {p4: {}, p5{}}, p2: {}, p3: {p4: {}, p5{}}}				
					typeToInterimTemplateMap[type] = result.template;
				}
					
				// templateStore.typeToRawTemplateMap[type] was deleted. However, raw template of this type needs to be restored when cycle is detected.
				// 1. If this type is not the startPoint, the resolved template for this type is not the final version. 
				//    Raw template is needed to work out this type's final version of template when this type becomes the start point in the future.
				// 2. If this type is the startPoint, the resolved template for this type is the final version. However,
				//    We may still need this type's raw template when other type's template is resolved.
				//    Example, {t1: {p1: t2}, t2: {p2: t1}}. When start point is t1, t1's resolved template is {p1: {p2: {}}. 
				//    When t2 becomes the start point later, we need the help from t1's raw template to work out {p2: {p1: {}} as t2's resolved template.
				//    If we use t1's resolved template, we will end up with {p2: {p1: {p2: {}}} as t2's resolved template, which is not good.
				
				// In the example of {t1: {p1: t2, p2: t1, p3: t2}, t2: {p4: t2, p5: t1}}, t2's raw template is restored when p1 is completed. But it would be used when p3 is handled.
				// Following order must be followed when it's time to resolve template: templateStore.typeToResolvedTemplateMap, typeToInterimTemplateMap, rawTemplate
				// templateStore.typeToCycledTemplateMap is not used, because the template there is in favour of its owner type.
				templateStore.typeToRawTemplateMap[type] = rawTemplate;
			} else {
				// no cycle. If isStartPoint is true, we get the template we plan to resolve. If isStartPoint is false, we get a great by-product
				templateStore.typeToResolvedTemplateMap[type] = result.template;
			}
		}
		
		return result;
	}
	
	return templateStore;
});
